@@ -19,6 +19,8 @@
 package org.apache.felix.framework.cache;
 
 import java.io.*;
+import java.nio.channels.FileChannel;
+import java.nio.channels.FileLock;
 import java.util.*;
 
 import org.apache.felix.framework.Logger;
@@ -51,6 +53,11 @@ import org.osgi.framework.Constants;
  *       property is set, then it will be calculated as being relative to
  *       the specified root directory.
  *   </li>
+ *   <li><tt>felix.cache.locking</tt> - Enables or disables bundle cache locking,
+ *       which is used to prevent concurrent access to the bundle cache. This is
+ *       enabled by default, but on older/smaller JVMs file channel locking is
+ *       not available; set this property to <tt>false</tt> to disable it.
+ *   </li>
  *   <li><tt>felix.cache.bufsize</tt> - Sets the buffer size to be used by
  *       the cache; the default value is 4096. The integer value of this
  *       string provides control over the size of the internal buffer of the
@@ -66,23 +73,108 @@ public class BundleCache
 {
     public static final String CACHE_BUFSIZE_PROP = "felix.cache.bufsize";
     public static final String CACHE_ROOTDIR_PROP = "felix.cache.rootdir";
+    public static final String CACHE_LOCKING_PROP = "felix.cache.locking";
     // TODO: CACHE - Remove this once we migrate the cache format.
     public static final String CACHE_SINGLEBUNDLEFILE_PROP = "felix.cache.singlebundlefile";
 
     protected static transient int BUFSIZE = 4096;
     protected static transient final String CACHE_DIR_NAME = "felix-cache";
     protected static transient final String CACHE_ROOTDIR_DEFAULT = ".";
+    protected static transient final String CACHE_LOCK_NAME = "cache.lock";
     protected static transient final String BUNDLE_DIR_PREFIX = "bundle";
 
     private static final SecureAction m_secureAction = new SecureAction();
 
     private final Logger m_logger;
     private final Map m_configMap;
+    private final Object m_lock;
 
     public BundleCache(Logger logger, Map configMap)
+        throws Exception
     {
         m_logger = logger;
         m_configMap = configMap;
+
+        // Create the cache directory, if it does not exist.
+        File cacheDir = determineCacheDir(m_configMap);
+        if (!getSecureAction().fileExists(cacheDir))
+        {
+            if (!getSecureAction().mkdirs(cacheDir))
+            {
+                m_logger.log(
+                    Logger.LOG_ERROR,
+                    "Unable to create cache directory: " + cacheDir);
+                throw new RuntimeException("Unable to create cache directory.");
+            }
+        }
+
+        Object locking = m_configMap.get(CACHE_LOCKING_PROP);
+        locking = (locking == null)
+            ? Boolean.TRUE.toString()
+            : locking.toString().toLowerCase();
+        if (((String) locking).equals(Boolean.TRUE.toString()))
+        {
+            File lockFile = new File(cacheDir, CACHE_LOCK_NAME);
+            FileChannel fc = null;
+            FileOutputStream fos = null;
+            try
+            {
+                if (!getSecureAction().fileExists(lockFile))
+                {
+                    fos = getSecureAction().getFileOutputStream(lockFile);
+                    fc = fos.getChannel();
+                }
+                else
+                {
+                    fos = getSecureAction().getFileOutputStream(lockFile);
+                    fc = fos.getChannel();
+                }
+            }
+            catch (Exception ex)
+            {
+                try
+                {
+                    if (fos != null) fos.close();
+                    if (fc != null) fc.close();
+                }
+                catch (Exception ex2)
+                {
+                    // Ignore.
+                }
+                throw new Exception("Unable to create bundle cache lock file: " + ex);
+            }
+            try
+            {
+                m_lock = fc.tryLock();
+            }
+            catch (Exception ex)
+            {
+                throw new Exception("Unable to lock bundle cache: " + ex);
+            }
+        }
+        else
+        {
+            m_lock = null;
+        }
+    }
+
+    public synchronized void release()
+    {
+        if (m_lock != null)
+        {
+            try
+            {
+                ((FileLock) m_lock).release();
+                ((FileLock) m_lock).channel().close();
+            }
+            catch (Exception ex)
+            {
+                // Not much we can do here, just log it.
+                m_logger.log(
+                    Logger.LOG_WARNING,
+                    "Exception releasing bundle cache.", ex);
+            }
+        }
     }
 
     /* package */ static SecureAction getSecureAction()
@@ -114,20 +206,8 @@ public class BundleCache
             // Use the default value.
         }
 
-        // Create the cache directory, if it does not exist.
-        File cacheDir = determineCacheDir(m_configMap);
-        if (!getSecureAction().fileExists(cacheDir))
-        {
-            if (!getSecureAction().mkdirs(cacheDir))
-            {
-                m_logger.log(
-                    Logger.LOG_ERROR,
-                    "Unable to create cache directory: " + cacheDir);
-                throw new RuntimeException("Unable to create cache directory.");
-            }
-        }
-
         // Create the existing bundle archives in the directory, if any exist.
+        File cacheDir = determineCacheDir(m_configMap);
         List archiveList = new ArrayList();
         File[] children = getSecureAction().listDirectory(cacheDir);
         for (int i = 0; (children != null) && (i < children.length); i++)
@@ -327,4 +407,4 @@ public class BundleCache
 
         return getSecureAction().deleteFile(target);
     }
-}
\ No newline at end of file
+}
