@@ -34,7 +34,6 @@ import org.apache.felix.framework.capabilityset.Capability;
 import org.apache.felix.framework.capabilityset.CapabilitySet;
 import org.apache.felix.framework.capabilityset.Directive;
 import org.apache.felix.framework.capabilityset.Requirement;
-import org.apache.felix.framework.util.Util;
 import org.apache.felix.framework.util.manifestparser.RequirementImpl;
 import org.osgi.framework.Bundle;
 import org.osgi.framework.Constants;
@@ -43,9 +42,6 @@ public class ResolverImpl implements Resolver
 {
     private final Logger m_logger;
 
-    // Reusable empty array.
-    private static final List<Wire> m_emptyWires = Util.m_emptyList;
-
     // Holds candidate permutations based on permutating "uses" chains.
     // These permutations are given higher priority.
     private final List<Map<Requirement, Set<Capability>>> m_usesPermutations =
@@ -79,14 +75,11 @@ public class ResolverImpl implements Resolver
 
                 ResolveException rethrow = null;
 
-                Map<Capability, Set<Requirement>> capDepSet = new HashMap();
-
                 do
                 {
                     rethrow = null;
 
                     modulePkgMap.clear();
-                    capDepSet.clear();
                     m_packageSourcesCache.clear();
 
                     candidateMap = (m_usesPermutations.size() > 0)
@@ -96,7 +89,7 @@ public class ResolverImpl implements Resolver
 
                     calculatePackageSpaces(
                         module, candidateMap, modulePkgMap,
-                        capDepSet, new HashMap(), new HashSet());
+                        new HashMap(), new HashSet());
 //System.out.println("+++ PACKAGE SPACES START +++");
 //dumpModulePkgMap(modulePkgMap);
 //System.out.println("+++ PACKAGE SPACES END +++");
@@ -104,7 +97,7 @@ public class ResolverImpl implements Resolver
                     try
                     {
                         checkPackageSpaceConsistency(
-                            module, candidateMap, modulePkgMap, capDepSet, new HashMap());
+                            false, module, candidateMap, modulePkgMap, new HashMap());
                     }
                     catch (ResolveException ex)
                     {
@@ -164,14 +157,11 @@ public class ResolverImpl implements Resolver
 
                 ResolveException rethrow = null;
 
-                Map<Capability, Set<Requirement>> capDepSet = new HashMap();
-
                 do
                 {
                     rethrow = null;
 
                     modulePkgMap.clear();
-                    capDepSet.clear();
 
                     candidateMap = (m_usesPermutations.size() > 0)
                         ? m_usesPermutations.remove(0)
@@ -179,13 +169,12 @@ public class ResolverImpl implements Resolver
 
                     calculatePackageSpaces(
                         module, candidateMap, modulePkgMap,
-                        capDepSet, new HashMap(), new HashSet());
+                        new HashMap(), new HashSet());
 
                     try
                     {
                         checkPackageSpaceConsistency(
-                            module, candidateMap, modulePkgMap,
-                            capDepSet, new HashMap());
+                            true, module, candidateMap, modulePkgMap, new HashMap());
                     }
                     catch (ResolveException ex)
                     {
@@ -540,7 +529,6 @@ public class ResolverImpl implements Resolver
         Module module,
         Map<Requirement, Set<Capability>> candidateMap,
         Map<Module, Packages> modulePkgMap,
-        Map<Capability, Set<Requirement>> capDepSet,
         Map<Capability, List<Module>> usesCycleMap,
         Set<Module> cycle)
     {
@@ -554,6 +542,7 @@ public class ResolverImpl implements Resolver
         // the module is resolved or not.
         List<Requirement> reqs = new ArrayList();
         List<Capability> caps = new ArrayList();
+        boolean isDynamicImport = false;
         if (module.isResolved())
         {
             // Use wires to get actual requirements and satisfying capabilities.
@@ -579,6 +568,7 @@ public class ResolverImpl implements Resolver
                 Capability cap = candCaps.iterator().next();
                 reqs.add(req);
                 caps.add(cap);
+                isDynamicImport = true;
                 // Can only dynamically import one at a time, so break
                 // out of the loop after the first.
                 break;
@@ -602,36 +592,62 @@ public class ResolverImpl implements Resolver
             }
         }
 
-        // First, add all exported packages to our package space.
-        calculateExportedPackages(module, modulePkgMap);
+        // First, add all exported packages to the target module's package space.
+        calculateExportedPackages(module, candidateMap, modulePkgMap);
         Packages modulePkgs = modulePkgMap.get(module);
 
-        // Second, add all imported packages to our candidate space.
+        // Second, add all imported packages to the target module's package space.
         for (int i = 0; i < reqs.size(); i++)
         {
             Requirement req = reqs.get(i);
             Capability cap = caps.get(i);
-            calculateExportedPackages(cap.getModule(), modulePkgMap);
+            calculateExportedPackages(cap.getModule(), candidateMap, modulePkgMap);
             mergeCandidatePackages(module, req, cap, modulePkgMap, candidateMap);
-            addCapabilityDependency(cap, req, capDepSet);
         }
 
-        // Third, ask our candidates to calculate their package spaces.
+        // Third, have all candidates to calculate their package spaces.
         for (int i = 0; i < caps.size(); i++)
         {
             calculatePackageSpaces(
                 caps.get(i).getModule(), candidateMap, modulePkgMap,
-                capDepSet, usesCycleMap, cycle);
+                usesCycleMap, cycle);
         }
 
-        // Fourth, add all of the uses constraints implied by our imported
-        // and required packages.
-        for (Entry<String, List<Blame>> entry : modulePkgs.m_importedPkgs.entrySet())
+        // Fourth, if the target module is unresolved or is dynamically importing,
+        // then add all the uses constraints implied by its imported and required
+        // packages to its package space.
+        // NOTE: We do not need to do this for resolved modules because their
+        // package space is consistent by definition and these uses constraints
+        // are only needed to verify the consistency of a resolving module. The
+        // only exception is if a resolve module is dynamically importing, then
+        // we need to calculate its uses constraints again to make sure the new
+        // import is consistent with the existing package space.
+        if (!module.isResolved() || isDynamicImport)
         {
-            for (Blame blame : entry.getValue())
+            for (Entry<String, List<Blame>> entry : modulePkgs.m_importedPkgs.entrySet())
             {
-                // Ignore modules that import from themselves.
-                if (!blame.m_cap.getModule().equals(module))
+                for (Blame blame : entry.getValue())
+                {
+                    // Ignore modules that import from themselves.
+                    if (!blame.m_cap.getModule().equals(module))
+                    {
+                        List<Requirement> blameReqs = new ArrayList();
+                        blameReqs.add(blame.m_reqs.get(0));
+
+                        mergeUses(
+                            module,
+                            modulePkgs,
+                            blame.m_cap,
+                            blameReqs,
+                            modulePkgMap,
+                            candidateMap,
+                            usesCycleMap);
+                    }
+                }
+            }
+            for (Entry<String, List<Blame>> entry : modulePkgs.m_requiredPkgs.entrySet())
+            {
+                for (Blame blame : entry.getValue())
                 {
                     List<Requirement> blameReqs = new ArrayList();
                     blameReqs.add(blame.m_reqs.get(0));
@@ -647,23 +663,6 @@ public class ResolverImpl implements Resolver
                 }
             }
         }
-        for (Entry<String, List<Blame>> entry : modulePkgs.m_requiredPkgs.entrySet())
-        {
-            for (Blame blame : entry.getValue())
-            {
-                List<Requirement> blameReqs = new ArrayList();
-                blameReqs.add(blame.m_reqs.get(0));
-
-                mergeUses(
-                    module,
-                    modulePkgs,
-                    blame.m_cap,
-                    blameReqs,
-                    modulePkgMap,
-                    candidateMap,
-                    usesCycleMap);
-            }
-        }
     }
 
     private void mergeCandidatePackages(
@@ -678,7 +677,8 @@ public class ResolverImpl implements Resolver
         else if (candCap.getNamespace().equals(Capability.MODULE_NAMESPACE))
         {
 // TODO: FELIX3 - THIS NEXT LINE IS A HACK. IMPROVE HOW/WHEN WE CALCULATE EXPORTS.
-            calculateExportedPackages(candCap.getModule(), modulePkgMap);
+            calculateExportedPackages(
+                candCap.getModule(), candidateMap, modulePkgMap);
 
             // Get the candidate's package space to determine which packages
             // will be visible to the current module.
@@ -686,8 +686,6 @@ public class ResolverImpl implements Resolver
 
             // We have to merge all exported packages from the candidate,
             // since the current module requires it.
-// TODO: FELIX3 - If a module imports its exports, then imported exports should
-//       be reexported to requiring bundles.
             for (Entry<String, Blame> entry : candPkgs.m_exportedPkgs.entrySet())
             {
                 mergeCandidatePackage(
@@ -770,18 +768,6 @@ public class ResolverImpl implements Resolver
         }
     }
 
-    private static void addCapabilityDependency(
-        Capability cap, Requirement req, Map<Capability, Set<Requirement>> capDepSet)
-    {
-        Set<Requirement> reqs = capDepSet.get(cap);
-        if (reqs == null)
-        {
-            reqs = new HashSet();
-            capDepSet.put(cap, reqs);
-        }
-        reqs.add(req);
-    }
-
     private void mergeUses(
         Module current, Packages currentPkgs,
         Capability mergeCap, List<Requirement> blameReqs, Map<Module, Packages> modulePkgMap,
@@ -860,15 +846,16 @@ public class ResolverImpl implements Resolver
     }
 
     private void checkPackageSpaceConsistency(
-        Module module, Map<Requirement, Set<Capability>> candidateMap,
-        Map<Module, Packages> modulePkgMap, Map<Capability, Set<Requirement>> capDepSet,
+        boolean isDynamicImport, Module module,
+        Map<Requirement, Set<Capability>> candidateMap,
+        Map<Module, Packages> modulePkgMap,
         Map<Module, Object> resultCache)
     {
-        if (module.isResolved())
+        if (module.isResolved() && !isDynamicImport)
         {
             return;
         }
-        else if (resultCache.containsKey(module))
+        else if(resultCache.containsKey(module))
         {
             return;
         }
@@ -891,7 +878,7 @@ public class ResolverImpl implements Resolver
                     {
                         sourceBlame = blame;
                     }
-                    else if (!sourceBlame.m_cap.equals(blame.m_cap))
+                    else if (!sourceBlame.m_cap.getModule().equals(blame.m_cap.getModule()))
                     {
                         // Try to permutate the conflicting requirement.
                         permutate(candidateMap, blame.m_reqs.get(0), m_importPermutations);
@@ -1108,8 +1095,8 @@ public class ResolverImpl implements Resolver
                     try
                     {
                         checkPackageSpaceConsistency(
-                            importBlame.m_cap.getModule(), candidateMap, modulePkgMap,
-                            capDepSet, resultCache);
+                            false, importBlame.m_cap.getModule(), candidateMap,
+                            modulePkgMap, resultCache);
                     }
                     catch (ResolveException ex)
                     {
@@ -1177,7 +1164,9 @@ public class ResolverImpl implements Resolver
     }
 
     private static void calculateExportedPackages(
-        Module module, Map<Module, Packages> modulePkgMap)
+        Module module,
+        Map<Requirement, Set<Capability>> candidateMap,
+        Map<Module, Packages> modulePkgMap)
     {
         Packages packages = modulePkgMap.get(module);
         if (packages != null)
@@ -1186,40 +1175,58 @@ public class ResolverImpl implements Resolver
         }
         packages = new Packages(module);
 
-        List<Capability> caps = module.getCapabilities();
-
-        if (caps.size() > 0)
+        // Get all exported packages.
+        Map<String, Capability> exports =
+            new HashMap<String, Capability>(module.getCapabilities().size());
+        for (Capability cap : module.getCapabilities())
         {
-            // Grab all exported packages that are not also imported.
-            for (int i = 0; i < caps.size(); i++)
+            if (cap.getNamespace().equals(Capability.PACKAGE_NAMESPACE))
             {
-// TODO: FELIX3 - Assume if a module imports the same package it
-//       exports that the import will overlap the export.
-                if (caps.get(i).getNamespace().equals(Capability.PACKAGE_NAMESPACE)
-                    && !hasOverlappingImport(module, caps.get(i)))
+                exports.put(
+                    (String) cap.getAttribute(Capability.PACKAGE_ATTR).getValue(),
+                    cap);
+            }
+        }
+        // Remove substitutable exports that were imported.
+        // For resolved modules look at the wires, for resolving
+        // modules look in the candidate map to determine which
+        // exports are substitutable.
+        if (module.isResolved())
+        {
+            for (Wire wire : module.getWires())
+            {
+                if (wire.getRequirement().getNamespace().equals(Capability.PACKAGE_NAMESPACE))
                 {
-                    packages.m_exportedPkgs.put(
-                        (String) caps.get(i).getAttribute(Capability.PACKAGE_ATTR).getValue(),
-                        new Blame(caps.get(i), null));
+                    String pkgName = (String) wire.getCapability()
+                        .getAttribute(Capability.PACKAGE_ATTR).getValue();
+                    exports.remove(pkgName);
                 }
             }
         }
-
-        modulePkgMap.put(module, packages);
-    }
-
-    private static boolean hasOverlappingImport(Module module, Capability cap)
-    {
-        List<Requirement> reqs = module.getRequirements();
-        for (int i = 0; i < reqs.size(); i++)
+        else
         {
-            if (reqs.get(i).getNamespace().equals(Capability.PACKAGE_NAMESPACE)
-                && CapabilitySet.matches(cap, reqs.get(i).getFilter()))
+            for (Requirement req : module.getRequirements())
             {
-                return true;
+                if (req.getNamespace().equals(Capability.PACKAGE_NAMESPACE))
+                {
+                    Set<Capability> cands = candidateMap.get(req);
+                    if ((cands != null) && !cands.isEmpty())
+                    {
+                        String pkgName = (String) cands.iterator().next()
+                            .getAttribute(Capability.PACKAGE_ATTR).getValue();
+                        exports.remove(pkgName);
+                    }
+                }
             }
         }
-        return false;
+        // Add all non-substituted exports to the module's package space.
+        for (Entry<String, Capability> entry : exports.entrySet())
+        {
+            packages.m_exportedPkgs.put(
+                entry.getKey(), new Blame(entry.getValue(), null));
+        }
+
+        modulePkgMap.put(module, packages);
     }
 
     private boolean isCompatible(
@@ -1263,7 +1270,7 @@ public class ResolverImpl implements Resolver
             return sources;
         }
 
-        return Util.m_emptyList;
+        return Collections.EMPTY_LIST;
     }
 
     private static List<Capability> getPackageSourcesInternal(
@@ -1328,7 +1335,7 @@ public class ResolverImpl implements Resolver
     {
         if (!module.isResolved() && !wireMap.containsKey(module))
         {
-            wireMap.put(module, m_emptyWires);
+            wireMap.put(module, (List<Wire>) Collections.EMPTY_LIST);
 
             List<Wire> packageWires = new ArrayList<Wire>();
             List<Wire> moduleWires = new ArrayList<Wire>();
@@ -1379,7 +1386,7 @@ public class ResolverImpl implements Resolver
         Module module, String pkgName, Map<Module, Packages> modulePkgMap,
         Map<Module, List<Wire>> wireMap, Map<Requirement, Set<Capability>> candidateMap)
     {
-        wireMap.put(module, m_emptyWires);
+        wireMap.put(module, (List<Wire>) Collections.EMPTY_LIST);
 
         List<Wire> packageWires = new ArrayList<Wire>();
 
